
#ifndef _hole_h
#define _hole_h

/* try removing all these.. well, need vecs etc  */ 
#include "sysstuff.h"
#include "string2.h"
#include "nrnb.h"
#include "network.h"
#include "confio.h"
#include "binio.h"
#include "copyrite.h"
#include "smalloc.h"
#include "main.h"
#include "pbc.h"
#include "force.h"
#include "macros.h"
#include "names.h"
#include "fatal.h"
#include "txtdump.h"
#include "futil.h"
#include "typedefs.h"
#include "random.h"
#include "vec.h"
#include "filenm.h"
#include "statutil.h"
#include "tgroup.h"
#include "pdebug.h"
#include "vcm.h"
#include "ebin.h"
#include "mdebin.h"
#include "dummies.h"
#include "physics.h" 
#include "pppm.h"
#include "calcmu.h"

/* now can declare and use rvecs, do iprods etc ... */

/* information for hole making routine */
typedef struct {
  rvec c;                 /* coordinates of center of hole                    */  
  real r;                 /* radius of hole                                   */  
  real lf;                /* force constant for cylindrical and perpendicular */  
  real sf;                /* repulsive forces                                 */  
  atom_id  lip1,lip2;     /* begin, end of lipid atoms                        */  
  atom_id  sol1,sol2;     /* begin, end of solvent atoms                      */  
  real z1,z2;             /* middle, lower and upper boundary of bilayer      */  
  bool cyl;               /* by grs ; if true use cylinder mode */
  bool grasp_ascii;       /* by grs ; if true use grasp ascii surf file */
  bool grasp_bin;         /* by grs ; if true use grasp binary surf file */
  bool msms_ascii;        /* by grs ; if true use msms surf files */
  bool debug_surf;        /* by grs ; will print a pdb showing forces and stop */
  bool resforces;         /* by grs ; average surface forces over all atoms (with force != 0) in residue */
  real sofs ;             /* by grs ; offset from molecular surface, eg 0.14 makes into a solvent accessible surface */
  int  supf ;             /* by grs ; frequency of updating the atom--mol. surface nn list (N^2 operation-yuk) */
  char *surffnm;            /* by grs ; file name for ASCII-format surface file (grasp or msms) */
  char *outfnm;            /* by grs ; file name for output information about number and depth of atoms within surface file */
  FILE *outfp;             /* by grs ; file pointer for outfnm file */ 
  int  type  ;             /* what type of run: Cyl, Grasp_bin, Grasp_ascii or MSMS */ 

} t_hole;

/* prototypes */

extern void readmsms_ascii ( char *filenm, int natom);
extern void readgrasp_ascii ( char *filenm, int natom);
extern void readgrasp_bin ( char *filenm, int natom);
extern void build_grasp_nn_list(int firstatom, int lastatom, rvec coord[],int toupd);
extern real get_grasp_surf_force(int ia, rvec coor, real offset, rvec force);
extern void average_surface_forces_over_residues(t_atoms *atoms, rvec cfarr[], 
					  int start, int homenr); 
extern void project_last_residues_forces(int nat, rvec avcf, rvec cfarr[], 
				  int firstat, int n); 
extern void free_grasp_surf_arrays(void); 
extern void output_surf_as_pdb (void); 
extern void readhole(t_hole *hole, char *hole_param_file) ; 
extern void inithole(t_hole *hole, int natoms) ;

#endif	/* _hole_h */
